package eu.jonahbauer.raytracing.math;

import org.jetbrains.annotations.NotNull;

public record Matrix3(
        double a11, double a12, double a13,
        double a21, double a22, double a23,
        double a31, double a32, double a33
) {

    public Matrix3() {
        this(1, 1, 1);
    }

    public Matrix3(double a11, double a22, double a33) {
        this(a11, 0, 0, 0, a22, 0, 0, 0, a33);
    }

    public @NotNull Matrix3 times(@NotNull Matrix3 other) {
        return new Matrix3(
                a11 * other.a11 + a12 * other.a21 + a13 * other.a31,
                a11 * other.a12 + a12 * other.a22 + a13 * other.a32,
                a11 * other.a13 + a12 * other.a23 + a13 * other.a33,
                a21 * other.a11 + a22 * other.a21 + a23 * other.a31,
                a21 * other.a12 + a22 * other.a22 + a23 * other.a32,
                a21 * other.a13 + a22 * other.a23 + a23 * other.a33,
                a31 * other.a11 + a32 * other.a21 + a33 * other.a31,
                a31 * other.a12 + a32 * other.a22 + a33 * other.a32,
                a31 * other.a13 + a32 * other.a23 + a33 * other.a33
        );
    }

    public @NotNull Matrix3 times(double other) {
        return new Matrix3(
                a11 * other, a12 * other, a13 * other,
                a21 * other, a22 * other, a23 * other,
                a31 * other, a32 * other, a33 * other
        );
    }

    public @NotNull Vec3 times(@NotNull Vec3 other) {
        return new Vec3(
                a11 * other.x() + a12 * other.y() + a13 * other.z(),
                a21 * other.x() + a22 * other.y() + a23 * other.z(),
                a31 * other.x() + a32 * other.y() + a33 * other.z()
        );
    }

    public @NotNull Matrix3 plus(@NotNull Matrix3 other) {
        return new Matrix3(
                a11 + other.a11, a12 + other.a12, a13 + other.a13,
                a21 + other.a21, a22 + other.a22, a23 + other.a23,
                a31 + other.a31, a32 + other.a32, a33 + other.a33
        );
    }

    public double det() {
        return a11 * a22 * a33 + a12 * a23 * a31 + a13 * a21 * a32
                - a13 * a22 * a31 - a23 * a32 * a11 - a33 * a12 * a21;
    }

    public @NotNull Matrix3 invert() {
        var det = det();
        if (det == 0) throw new IllegalStateException();
        var t = 1 / det;
        return new Matrix3(
                t * (a22 * a33 - a23 * a32), t * (-a12 * a33 + a13 * a32), t * (a12 * a23 - a13 * a22),
                t * (-a21 * a33 + a23 * a31), t * (a11 * a33 - a13 * a31), t * (-a11 * a23 + a13 * a21),
                t * (a21 * a32 - a22 * a31), t * (a11 * a32 + a12 * a31), t * (a11 * a22 - a12 * a21)
        );
    }
}
