package eu.jonahbauer.raytracing.render.spectral;

import eu.jonahbauer.raytracing.render.spectral.colors.ColorXYZ;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectra;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;

/**
 * A set of sampled wavelength that can be tracked together.
 */
public final class SampledWavelengths {
    private final double @NotNull[] lambdas;
    private final double @NotNull[] pdf;

    public static @NotNull SampledWavelengths uniform(double rng) {
        return uniform(rng, Spectrum.LAMBDA_MIN, Spectrum.LAMBDA_MAX);
    }

    public static @NotNull SampledWavelengths uniform(double rng, double min, double max) {
        var lambdas = new double[Spectrum.SAMPLES];

        // choose first sample at random
        lambdas[0] = (1 - rng) * min + rng * max;

        // choose next samples in equal intervals, wrapping if necessary
        var delta = (max - min) / Spectrum.SAMPLES;
        for (int i = 1; i < Spectrum.SAMPLES; i++) {
            lambdas[i] = lambdas[i - 1] + delta;
            if (lambdas[i] > max) {
                lambdas[i] = min + (lambdas[i] - max);
            }
        }

        var pdf = new double[Spectrum.SAMPLES];
        Arrays.fill(pdf, 1 / (max - min));
        return new SampledWavelengths(lambdas, pdf);
    }

    private SampledWavelengths(double @NotNull[] lambdas, double @NotNull[] pdf) {
        this.lambdas = lambdas;
        this.pdf = pdf;
    }

    public double get(int index) {
        return lambdas[index];
    }

    public int size() {
        return lambdas.length;
    }

    /**
     * Terminates the secondary wavelengths. This method should be called after a wavelength-dependent operation like
     * refraction that makes it incorrect to track multiple wavelengths together.
     */
    public void terminateSecondary() {
        if (pdf.length < 2 || pdf[1] == 0) return;
        Arrays.fill(pdf, 1, pdf.length, 0d);
        pdf[0] /= pdf.length;
    }

    @NotNull
    ColorXYZ toXYZ(@NotNull SampledSpectrum spectrum) {
        var x = Spectra.X.sample(this);
        var y = Spectra.Y.sample(this);
        var z = Spectra.Z.sample(this);

        return new ColorXYZ(
                toXYZ0(spectrum, x) / ColorXYZ.CIE_Y_INTEGRAL,
                toXYZ0(spectrum, y) / ColorXYZ.CIE_Y_INTEGRAL,
                toXYZ0(spectrum, z) / ColorXYZ.CIE_Y_INTEGRAL
        );
    }

    private double toXYZ0(@NotNull SampledSpectrum spectrum, @NotNull SampledSpectrum cie) {
        var avg = 0d;
        for (int i = 0; i < spectrum.size(); i++) {
            var pdf = this.pdf[i];
            double value;
            if (pdf == 0) {
                value = 0;
            } else {
                value = spectrum.get(i) * cie.get(i) / pdf;
            }
            avg = Math.fma(1d / (i + 1), value - avg, avg);
        }
        return avg;
    }
}
