package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.math.Matrix3;
import eu.jonahbauer.raytracing.render.spectral.spectrum.DenselySampledSpectrum;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;
import eu.jonahbauer.raytracing.render.texture.Color;
import org.jetbrains.annotations.NotNull;

import java.util.NoSuchElementException;
import java.util.Objects;

/**
 * An RGB color space.
 */
public final class ColorSpace {
    private final @NotNull Chromaticity r;
    private final @NotNull Chromaticity g;
    private final @NotNull Chromaticity b;
    private final @NotNull Chromaticity w;
    private final @NotNull DenselySampledSpectrum illuminant;

    private final @NotNull Matrix3 XYZfromRGB;
    private final @NotNull Matrix3 RGBfromXYZ;
    private final @NotNull SpectrumTable RGBtoSpectrumTable;

    public ColorSpace(
            @NotNull Chromaticity r, @NotNull Chromaticity g, @NotNull Chromaticity b,
            @NotNull Spectrum illuminant, @NotNull SpectrumTable table
    ) {
        this.r = Objects.requireNonNull(r, "r");
        this.g = Objects.requireNonNull(r, "g");
        this.b = Objects.requireNonNull(r, "b");
        this.illuminant = new DenselySampledSpectrum(illuminant);
        this.RGBtoSpectrumTable = table;

        var W = illuminant.toXYZ();
        this.w = W.xy();

        var R = new ColorXYZ(r);
        var G = new ColorXYZ(g);
        var B = new ColorXYZ(b);
        var rgb = new Matrix3(
                R.x(), G.x(), B.x(),
                R.y(), G.y(), B.y(),
                R.z(), G.z(), B.z()
        );
        var C = rgb.invert().times(W.toVec3());

        this.XYZfromRGB = rgb.times(new Matrix3(C.x(), C.y(), C.z()));
        this.RGBfromXYZ = XYZfromRGB.invert();
    }

    public @NotNull Color toRGB(@NotNull ColorXYZ xyz) {
        var out = RGBfromXYZ.times(xyz.toVec3());
        return new Color(out.x(), out.y(), out.z());
    }

    public @NotNull ColorXYZ toXYZ(@NotNull Color rgb) {
        var out = XYZfromRGB.times(rgb.toVec3());
        return new ColorXYZ(out);
    }

    public @NotNull SigmoidPolynomial toSpectrum(@NotNull Color rgb) {
        return RGBtoSpectrumTable.get(new Color(
                Math.max(0, rgb.r()),
                Math.max(0, rgb.g()),
                Math.max(0, rgb.b())
        ));
    }

    public @NotNull Chromaticity r() {
        return r;
    }

    public @NotNull Chromaticity g() {
        return g;
    }

    public @NotNull Chromaticity b() {
        return b;
    }

    public @NotNull Chromaticity w() {
        return w;
    }

    public @NotNull DenselySampledSpectrum illuminant() {
        return illuminant;
    }
}
