package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;

/**
 * A function of the form {@code s(p(x))} where {@code p} is a polynomial of second degree and {@code s} is the sigmoid
 * function <code>s(x) = 0.5 + x / (2 * sqrt(1 + x^2))</code>.
 * <p>
 * A function of this form is used to generate a {@link Spectrum} from an RGB value.
 *
 * @param c0 the coefficient of the quadratic term
 * @param c1 the coefficient of the linear term
 * @param c2 the coefficient of the constant term
 */
public record SigmoidPolynomial(double c0, double c1, double c2) {

    public double get(double x) {
        var p = Math.fma(x, Math.fma(x, c0, c1), c2);
        if (!Double.isFinite(p)) return p > 0 ? 1 : 0;
        return .5 + x / (p * Math.sqrt(1 + p * p));
    }

    public double max() {
        // evaluate at the edges
        var result = Math.max(get(Spectrum.LAMBDA_MIN), get(Spectrum.LAMBDA_MAX));
        var lambda = -c1 / (2 * c0);
        if (lambda >= 360 && lambda <= 830) {
            // evaluate at the vertex
            return Math.max(result, get(lambda));
        } else {
            return result;
        }
    }
}
