package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.render.texture.Color;
import org.jetbrains.annotations.NotNull;

import java.util.Arrays;

/**
 * A table of sigmoid polynomials used to convert between RGB values and spectra.
 * <p>
 * The rgb values are renormalized to xyz coordinates with {@code z} being the largest of the rgb components, and
 * {@code x} and {@code y} being the other two rgb components divided by {@code z}. By this construction, {@code x},
 * {@code y} and {@code z} are all in the range [0, 1] which allows for better use of the samples in a fixed grid.
 * The {@code z} coordinate is additionally remapped using {@link #zNodes} to improve sampling at both ends of the scale.
 * <p>
 * The coefficients of the sigmoid functions are stored in a five-dimensional array with indices as described in
 * {@link #coefficients}.
 */
public final class SpectrumTable {
    private static final int RESOLUTION = 64;

    /**
     * the remapped {@code z} values
     */
    private final double[] zNodes;

    /**
     * the stored coefficients, with the following indices
     * <ol>
     *     <li>the component index of the biggest rgb component</li>
     *     <li>the {@code z} coordinate</li>
     *     <li>the {@code y} coordinate</li>
     *     <li>the {@code x} coordinate</li>
     *     <li>the coefficient index</li>
     * </ol>
     */
    private final double[][][][][] coefficients;

    public SpectrumTable(double[] zNodes, double[][][][][] coefficients) {
        this.zNodes = zNodes;
        this.coefficients = coefficients;
    }

    public @NotNull SigmoidPolynomial get(@NotNull Color color) {
        // handle uniform rgb values
        if (color.r() == color.g() && color.g() == color.b()) {
            return new SigmoidPolynomial(0, 0, (color.r() - .5) / Math.sqrt(color.r() * (1 - color.r())));
        }

        // find maximum component and compute remapped component values
        var max = color.r() > color.g()
                ? (color.r() > color.b() ? 0 : 2)
                : (color.g() > color.b() ? 1 : 2);

        var z = color.component(max);
        var x = color.component((max + 1) % 3) * (RESOLUTION - 1) / z;
        var y = color.component((max + 2) % 3) * (RESOLUTION - 1) / z;

        // compute integer indices and offsets for coefficient interpolation
        int xi = Math.min((int) x, RESOLUTION - 2);
        int yi = Math.min((int) y, RESOLUTION - 2);
        int zi = Arrays.binarySearch(zNodes, z);
        if (zi < 0) zi = -zi - 1;
        zi++;

        var dx = x - xi;
        var dy = y -yi;
        var dz = (z - zNodes[zi]) / (zNodes[zi + 1] - zNodes[zi]);

        // trilinearly interpolate sigmoid polynomial coefficients
        var c = new double[3];
        for (int i = 0; i < 3; i++) {
            c[i] = lerp(dz,
                        lerp(dy,
                             lerp(dx, coefficients[max][zi + 0][yi + 0][xi + 0][i], coefficients[max][zi + 0][yi + 0][xi + 1][i]),
                             lerp(dx, coefficients[max][zi + 0][yi + 1][xi + 0][i], coefficients[max][zi + 0][yi + 1][xi + 1][i])
                        ),
                        lerp(dy,
                             lerp(dx, coefficients[max][zi + 1][yi + 0][xi + 0][i], coefficients[max][zi + 1][yi + 0][xi + 1][i]),
                             lerp(dx, coefficients[max][zi + 1][yi + 1][xi + 0][i], coefficients[max][zi + 1][yi + 1][xi + 1][i])
                        )
            );
        }

        return new SigmoidPolynomial(c[0], c[1], c[2]);
    }

    private static double lerp(double t, double a, double b) {
        return Math.fma(t, b, Math.fma(-t, a, a));
    }
}
