package eu.jonahbauer.raytracing.render.spectral.spectrum;

import org.jetbrains.annotations.NotNull;

import java.util.Arrays;

/**
 * A spectrum sampled in one nanometer intervals.
 */
public final class DenselySampledSpectrum implements Spectrum {
    private final double[] samples;
    private final int min;

    private final double max;

    public DenselySampledSpectrum(@NotNull Spectrum spectrum) {
        this(spectrum, LAMBDA_MIN, LAMBDA_MAX);
    }

    public DenselySampledSpectrum(@NotNull Spectrum spectrum, int min, int max) {
        if (max - min + 1 <= 0) throw new IllegalArgumentException("samples must not be empty");
        this.samples = new double[max - min + 1];
        var maxValue = 0d;
        for (int lambda = min; lambda <= max; lambda++) {
            var sample = spectrum.get(lambda);
            if (sample > maxValue) maxValue = sample;
            this.samples[lambda] = sample;
        }
        this.min = min;
        this.max = maxValue;
    }

    public DenselySampledSpectrum(double @NotNull[] samples, int lambdaMin) {
        if (samples.length == 0) throw new IllegalArgumentException("samples must not be empty");
        this.samples = Arrays.copyOf(samples, samples.length);
        this.min = lambdaMin;
        this.max = Arrays.stream(this.samples).max().orElseThrow();
    }

    @Override
    public double max() {
        return max;
    }

    @Override
    public double get(double lambda) {
        int offset = (int) Math.round(lambda) - min;
        if (offset < 0 || offset >= samples.length) return 0;
        return samples[offset];
    }
}
