package eu.jonahbauer.raytracing.render.spectral.spectrum;

import org.jetbrains.annotations.NotNull;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UncheckedIOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

public final class Spectra {
    /**
     * the CIE XYZ color matching curve for X
     */
    public static final Spectrum X;
    /**
     * the CIE XYZ color matching curve for Y
     */
    public static final Spectrum Y;
    /**
     * the CIE XYZ color matching curve for Z
     */
    public static final Spectrum Z;
    /**
     * the CIE standard illuminant D50
     */
    public static final Spectrum D50;
    /**
     * the CIE standard illuminant D65
     */
    public static final Spectrum D65;

    static {
        var xyz = read("/eu/jonahbauer/raytracing/spectrum/CIE_xyz_1931_2deg.csv");
        X = xyz[0];
        Y = xyz[1];
        Z = xyz[2];
        D50 = read("/eu/jonahbauer/raytracing/spectrum/CIE_std_illum_D50.csv")[0];
        D65 = read("/eu/jonahbauer/raytracing/spectrum/CIE_std_illum_D65.csv")[0];
    }

    private static @NotNull DenselySampledSpectrum[] read(@NotNull String path) {
        var lambda = new ArrayList<Double>();
        var values = new ArrayList<List<Double>>();

        try (
                var is = Spectra.class.getResourceAsStream(path);
                var in = new BufferedReader(new InputStreamReader(is, StandardCharsets.US_ASCII))
        ) {
            int count = -1;
            String line;
            while ((line = in.readLine()) != null) {
                var parts = line.split(",");
                if (count == -1) {
                    count = parts.length - 1;
                    for (int i = 0; i < count; i++) values.add(new ArrayList<>());
                } else if (count != parts.length - 1) {
                    throw new IllegalStateException("invalid spectrum file: " + path);
                }

                lambda.add(Double.parseDouble(parts[0]));
                for (int i = 0; i < count; i++) {
                    values.get(i).add(Double.parseDouble(parts[i + 1]));
                }
            }
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }

        var out = new DenselySampledSpectrum[values.size()];
        var lambdaArray = lambda.stream().mapToDouble(Double::doubleValue).toArray();
        for (int i = 0; i < values.size(); i++) {
            var valueArray = values.get(i).stream().mapToDouble(Double::doubleValue).toArray();
            var s = new PiecewiseLinearSpectrum(lambdaArray, valueArray);
            out[i] = new DenselySampledSpectrum(s);
        }

        return out;
    }

    private Spectra() {
        throw new UnsupportedOperationException();
    }
}
